!> @file modules.f90
!------------------------------------------------------------------------------!
! This file is part of PALM.
!
! PALM is free software: you can redistribute it and/or modify it under the
! terms of the GNU General Public License as published by the Free Software
! Foundation, either version 3 of the License, or (at your option) any later
! version.
!
! PALM is distributed in the hope that it will be useful, but WITHOUT ANY
! WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
! A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License along with
! PALM. If not, see <http://www.gnu.org/licenses/>.
!
! Copyright 1997-2017 Leibniz Universitaet Hannover
!------------------------------------------------------------------------------!
!
! Current revisions:
! ------------------
! 
! 
! Former revisions:
! -----------------
! $Id: modules.f90 2233 2017-05-30 18:08:54Z suehring $
!
! 2232 2017-05-30 17:47:52Z suehring
! Renamed wall_flags_0 and wall_flags_00 into advc_flags_1 and advc_flags_2, 
! respectively. Moreover, introduced further flag array wall_flags_0. 
!
! Adjustments for new topography concept:
!   -fwxm, fwxp, fwym, fwyp, fxm, fxp, fym, fyp, rif_wall, wall_e_x, wall_e_y, 
!   -wall_v, wall_u, wall_w_x, wall_w_y, wall_qflux, wall_sflux, wall_nrflux, 
!   -wall_qrflux
!
! Adjustments for new surface concept:
!   +land_surface
!   -z0, z0h, z0q, us, ts, qs, qsws, nrs, nrsws, qrs, qrsws, ssws, ss, saswsb 
!   -nzb_diff_u, nzb_diff_v, nzt_diff
!   -uswst, vswst, tswst, sswst, saswst, qswst, qrswst, nrswst, qswst_remote
!
! Generic tunnel setup:
!   +tunnel_height, tunnel_length, tunnel_width_x, tunnel_width_y, 
!   +tunnel_wall_depth
!
! Topography input via netcdf
!   +lod
! 
! 2200 2017-04-11 11:37:51Z suehring
! -monotonic_adjustment
! 
! 2174 2017-03-13 08:18:57Z maronga
! Changed default values for most_method to 'newton'
! 
! 2118 2017-01-17 16:38:49Z raasch
! -acc_rank, background_communication, i_left, i_right, j_south, j_north,
!  num_acc_per_node, on_device
! 
! 2107 2017-01-09 12:21:49Z kanani
! Preparation for doxygen comments (Giersch)
! 
! 2050 2016-11-08 15:00:55Z gronemeier
! Implement turbulent outflow condition
! 
! 2037 2016-10-26 11:15:40Z knoop
! Anelastic approximation implemented
! 
! 2031 2016-10-21 15:11:58Z knoop
! renamed variable rho to rho_ocean and rho_av to rho_ocean_av
! 
! 2011 2016-09-19 17:29:57Z kanani
! +urban_surface, +lsf_exception, +varnamelength
! 
! 2007 2016-08-24 15:47:17Z kanani
! Increased DIMENSION of data_output, data_output_user, do2d, do3d
! 
! 2000 2016-08-20 18:09:15Z knoop
! Forced header and separation lines into 80 columns
! 
! 1992 2016-08-12 15:14:59Z suehring
! +constant_top_scalarflux, top_scalarflux
! default of bc_s_t adjusted
! 
! 1968 2016-07-18 12:01:49Z suehring
! Changed dimension for MPI-datatypes type_x_int and type_y_int
! 
! 1960 2016-07-12 16:34:24Z suehring
! Separate humidity and passive scalar
! +bc_s_t_val, diss_s_s, diss_l_s, flux_s_s, flux_l_s, s, sp, s1, s2, s3, ssws_av, 
!  s_init, s_surf, sums_wsss_ws_l, ss, ssws, sswst, ts_m, wall_sflux
! +constant_scalarflux, ibc_s_b, ibc_s_t, s_vertical_gradient_level_ind
!
! Unused variables removed
! -gamma_x, gamma_y, gamma_z, var_x, var_y, var_z
!
! Change initial values (in order to unify gradient calculation): 
! pt_vertical_gradient_level, sa_vertical_gradient_level
!
! 1957 2016-07-07 10:43:48Z suehring
! +fl_max, num_leg, num_var_fl, num_var_fl_user, var_fl_max, virtual_flight
! 
! 1918 2016-05-27 14:35:57Z raasch
! default timestep switched from -1.0 to +1.0 in order to avoid wrong sign of
! initially calculated divergence
!
! 1906 2016-05-24 14:38:08Z suehring
! default value of mg_switch_to_pe0_level changed to -1
!
! 1849 2016-04-08 11:33:18Z hoffmann
! bfactor, mass_of_solute, molecular_weight_of_solute, molecular_weight_of_water,
! vanthoff moved to mod_particle_attributes.
! dt_micro and several cloud_parameters moved to microphysics_mod.
! 1d-microphysics profiles moved to microphysics_mod.
!
! 1845 2016-04-08 08:29:13Z raasch
! -nzb_2d
!
! 1833 2016-04-07 14:23:03Z raasch
! spectra parameter moved to spectra module
!
! 1831 2016-04-07 13:15:51Z hoffmann
! curvature_solution_effects removed
! turbulence renamed collision_turbulence, drizzle renamed 
! cloud_water_sedimentation
!
! 1822 2016-04-07 07:49:42Z hoffmann
! icloud_scheme removed. microphysics_sat_adjust, microphysics_kessler,
! microphysics_seifert added.
!
! 1815 2016-04-06 13:49:59Z raasch
! cpp-directive for decalpha removed
!
! 1808 2016-04-05 19:44:00Z raasch
! MPI module used by default on all machines
!
! 1804 2016-04-05 16:30:18Z maronga
! Removed code for parameter file check (__check)
! 
! 1788 2016-03-10 11:01:04Z maronga
! Added roughness length for moisture (z0q)
!
! 1786 2016-03-08 05:49:27Z raasch
! module spectrum moved to new separate module
!
! 1783 2016-03-06 18:36:17Z raasch
! netcdf variables moved to the netcdf-interface module
!
! 1779 2016-03-03 08:01:28Z raasch
! coupling_char extended to LEN=3
!
! 1764 2016-02-28 12:45:19Z raasch
! some reformatting
!
! 1762 2016-02-25 12:31:13Z hellstea
! +nest_* variables, size of volume_flow arrays increased by one element
!
! 1738 2015-12-18 13:56:05Z raasch
! +mean_surface_level_height
!
! 1695 2015-10-27 10:03:11Z maronga
! Removed rif (forgotten in last revision)
! 
! 1693 2015-10-27 08:35:45Z maronga
! Renamed zp -> z_mo
! 
! 1691 2015-10-26 16:17:44Z maronga
! Renamed Obukhov length. Added ol, removed rif. Increased number of profiles 
! (pr_palm). Changed default values for dissipation_1d to 'detering' and 
! (mixing_length_1d to 'blackadar'. Added most_method. rif_min and rif_max
! renamed to zeta_min and zeta_max and new values assigned.
! 
! 1682 2015-10-07 23:56:08Z knoop
! Code annotations made doxygen readable 
! 
! 1677 2015-10-02 13:25:23Z boeske
! +ngp_yz_int, type_xz_int, type_yz_int
! 
! 1666 2015-09-23 07:31:10Z raasch
! +user_interface_current_revision, user_interface_required_revision in
! control_parameters
!
! 1639 2015-08-31 14:46:48Z knoop
! Bugfix: string 'unknown' extended to match LEN=13 
!
! 1575 2015-03-27 09:56:27Z raasch
! +ngp_xz
!
! 1560 2015-03-06 10:48:54Z keck
! +recycling_yshift
! 
! 1557 2015-03-05 16:43:04Z suehring
! +monotonic_adjustment
! 
! 1551 2015-03-03 14:18:16Z maronga
! Increased pr_palm to 120. Increased length of dots_unit and dots_label to 13
! digits. Increased length of domask, do2d, and do3d to 20 digits.
! 
! 1496 2014-12-02 17:25:50Z maronga
! Renamed "radiation" -> "cloud_top_radiation"
! 
! 1484 2014-10-21 10:53:05Z kanani
! Changes due to new module structure of the plant canopy model:
!   canopy-model related parameters/variables moved to module 
!   plant_canopy_model_mod
! 
! 1468 2014-09-24 14:06:57Z maronga
! Adapted for use on up to 6-digit processor cores.
! Increased identifier string length for user-defined quantities to 20.
! 
! 1450 2014-08-21 07:31:51Z heinze
! ntnudge from 100 to 1000 increased to allow longer simulations
! 
! 1431 2014-07-15 14:47:17Z suehring
! +var_d
! 
! 1429 2014-07-15 12:53:45Z knoop
! +ensemble_member_nr to prepare the random_generator for ensemble runs
!
! 1382 2014-04-30 12:15:41Z boeske
! Renamed variables which store large scale forcing tendencies
! pt_lsa -> td_lsa_lpt, pt_subs -> td_sub_lpt, 
! q_lsa  -> td_lsa_q,   q_subs  -> td_sub_q
! 
! 1365 2014-04-22 15:03:56Z boeske
! Usage of large scale forcing enabled:
! increase pr_palm from 90 to 100 to allow for more standard profiles
! + ngp_sums_ls, pt_lsa, pt_subs, q_lsa, q_subs, tmp_tnudge, sums_ls_l, 
! use_subsidence_tendencies
! 
! 1361 2014-04-16 15:17:48Z hoffmann
! tend_* removed
! call_microphysics_at_all_substeps added
! default of drizzle set to true
! 
! 1359 2014-04-11 17:15:14Z hoffmann
! particle_attributes moved to mod_particle_attributes.f90
! 
! 1353 2014-04-08 15:21:23Z heinze
! REAL constants provided with KIND-attribute
! 
! 1327 2014-03-21 11:00:16Z raasch
! REAL constants defined as wp-kind
! -avs_output, data_output_format, do3d_compress, iso2d_output, netcdf_output
!
! 1320 2014-03-20 08:40:49Z raasch
! ONLY-attribute added to USE-statements,
! kind-parameters added to all INTEGER and REAL declaration statements,
! kinds are defined in new module kinds,
! old module precision_kind is removed,
! revision history before 2012 removed,
! comment fields (!:) to be used for variable explanations added to
! all variable declaration statements
!
! 1318 2014-03-17 13:35:16Z raasch
! module cpulog moved to new separate module-file
! interface for cpu_log removed
!
! 1314 2014-03-14 18:25:17Z suehring
! + log_z_z0, number_of_sublayers, z0_av_global 
! 1308 2014-03-13 14:58:42Z fricke
! +ntdim_2d_xy, ntdim_2d_xz, ntdim_2d_yz, ntdim_3d
!
! 1257 2013-11-08 15:18:40Z raasch
! set default values for grid indices of maximum velocity components
! u|v|w_max_ijk
!
! 1241 2013-10-30 11:36:58Z heinze
! Usage of nudging enabled 
! +nudging, ntnudge, ptnudge, qnudge, tnudge, unudge, vnudge, wnudge
! increase pr_palm from 80 to 90 to allow for more standard profiles
!
! Enable prescribed time depenend surface fluxes and geostrophic wind read in
! from external file LSF_DATA
! +large_scale_forcing, lsf_surf, lsf_vert, nlsf, time_surf, shf_surf, qsws_surf,
!  pt_surf, q_surf, p_surf, time_vert, ug_vert, vg_vert, wsubs_vert
!
! 1221 2013-09-10 08:59:13Z raasch
! wall_flags_0 changed to 32bit int, +wall_flags_00,
! +rflags_s_inner, rflags_invers
!
! 1216 2013-08-26 09:31:42Z raasch
! +transpose_compute_overlap,
! several variables are now defined in the serial (non-parallel) case also
!
! 1212 2013-08-15 08:46:27Z raasch
! +tri
!
! 1179 2013-06-14 05:57:58Z raasch
! +reference_state, ref_state, use_initial_profile_as_reference, vpt_reference,
! use_reference renamed use_single_reference_value
!
! 1159 2013-05-21 11:58:22Z fricke
! -bc_lr_dirneu, bc_lr_neudir, bc_ns_dirneu, bc_ns_neudir
! +use_cmax
!
! 1128 2013-04-12 06:19:32Z raasch
! +background_communication, i_left, i_right, j_north, j_south, req, req_count,
! send_receive, sendrecv_in_background, wait_stat
!
! 1115 2013-03-26 18:16:16Z hoffmann
! unused variables removed
!
! 1113 2013-03-10 02:48:14Z raasch
! +on_device
!
! 1111 2013-03-08 23:54:10Z raasch
! +tric, nr_timesteps_this_run
!
! 1106 2013-03-04 05:31:38Z raasch
! array_kind renamed precision_kind, pdims defined in serial code
! bugfix: default value assigned to coupling_start_time
!
! 1095 2013-02-03 02:21:01Z raasch
! FORTRAN error in r1092 removed
!
! 1092 2013-02-02 11:24:22Z raasch
! character length in some derived type changed for better alignment
!
! 1065 2012-11-22 17:42:36Z hoffmann
! + c_sedimentation, limiter_sedimentation, turbulence, a_1, a_2, a_3, b_1, b_2, 
! + b_3, c_1, c_2, c_3, beta_cc
!
! bottom boundary condition of qr, nr changed from Dirichlet to Neumann
!
! 1053 2012-11-13 17:11:03Z hoffmann
! necessary expansions according to the two new prognostic equations (nr, qr) 
! of the two-moment cloud physics scheme:
! +*_init, flux_l_*, diss_l_*, flux_s_*, diss_s_*, *sws, *swst, tend_*, *, *_p
! +t*_m, *_1, *_2, *_3, *_av, bc_*_b, bc_*_t, ibc_*_b, ibc_*_t, bc_*_t_val, 
! +*_vertical_gradient, *_surface_initial_change, *_vertical_gradient_level, 
! +*_vertical_gradient_level_ind, *_surface, constant_waterflux_*,  
! +cloud_scheme, icloud_scheme, surface_waterflux_*, sums_ws*s_ws_l, wall_*flux
!
! constants for the two-moment scheme:
! +a_vent, a_term, b_vent, b_term, c_evap, c_term, cof, eps_sb, k_cc, k_cr, k_rr, 
! +k_br, kappa_rr, kin_vis_air, mu_constant_value, nc, pirho_l, dpirho_l, rho_1,
! +schmidt, schmidt_p_1d3, stp, x0, xmin, xmax, dt_precipitation, w_precipitation
!
! steering parameters for the two_moment scheme:
! +mu_constant, ventilation_effect
!
! 1036 2012-10-22 13:43:42Z raasch
! code put under GPL (PALM 3.9)
!
! 1031 2012-10-19 14:35:30Z raasch
! +output_format_netcdf
!
! 1015 2012-09-27 09:23:24Z raasch
! +acc_rank, num_acc_per_node,
! -adjust_mixing_length
!
! 1010 2012-09-20 07:59:54Z raasch
! pointer free version can be generated with cpp switch __nopointer
!
! 1003 2012-09-14 14:35:53Z raasch
! -grid_matching, nxa, nya, etc., nnx_pe, nny_pe, spl_*
!
! 1001 2012-09-13 14:08:46Z raasch
! -asselin_filter_factor, cut_spline_overshoot, dt_changed, last_dt_change,
! last_dt_change_1d, long_filter_factor, overshoot_limit_*, ups_limit_*
! several pointer/target arrays converted to normal ones
!
! 996 2012-09-07 10:41:47Z raasch
! -use_prior_plot1d_parameters
!
! 978 2012-08-09 08:28:32Z fricke
! +c_u_m, c_u_m_l, c_v_m, c_v_m_l, c_w_m, c_w_m_l,
! +bc_lr_dirneu, bc_lr_neudir, bc_ns_dirneu, bc_ns_neudir
! -km_damp_x, km_damp_y, km_damp_max, outflow_damping_width
! +z0h, z0h_av, z0h_factor, z0h1d
! +ptdf_x, ptdf_y, pt_damping_width, pt_damping_factor 
!
! 964 2012-07-26 09:14:24Z raasch
! -cross_linecolors, cross_linestyles, cross_normalized_x, cross_normx_factor,
! cross_normalized_y, cross_normy_factor, cross_pnc_local,
! cross_profile_numbers, cross_profile_number_counter, cross_uxmax,
! cross_uxmax_computed, cross_uxmax_normalized,
! cross_uxmax_normalized_computed, cross_uxmin, cross_uxmin_computed,
! cross_uxmin_normalized, cross_uxmin_normalized_computed, cross_uymax,
! cross_uymin, cross_xtext, dopr_crossindex, dopr_label, linecolors, linestyles,
! nz_do1d, profil_output, z_max_do1d, z_max_do1d_normalized
!
! 951 2012-07-19 14:22:52Z hoffmann
! changing profile_columns and profile_rows
!
! 940 2012-07-09 14:31:00Z raasch
! +neutral
!
! 927 2012-06-06 19:15:04Z raasch
! +masking_method
!
! 880 2012-04-13 06:28:59Z raasch
! gathered_size, subdomain_size moved to control_parameters
!
! 866 2012-03-28 06:44:41Z raasch
! interface for global_min_max changed
!
! 861 2012-03-26 14:18:34Z suehring
! +wall_flags_0
! -boundary_flags
! +nzb_max
! +adv_sca_1, +adv_mom_1
!
! 849 2012-03-15 10:35:09Z raasch
! +deleted_particles, deleted_tails, tr.._count_sum, tr.._count_recv_sum in
! particle_attributes,
! +de_dx, de_dy, de_dz in arrays_3d,
! first_call_advec_particles renamed first_call_lpm
!
! 828 2012-02-21 12:00:36Z raasch
! +dissipation_classes, radius_classes, use_kernel_tables,
! particle feature color renamed class
!
! 825 2012-02-19 03:03:44Z raasch
! +bfactor, curvature_solution_effects, eps_ros, molecular_weight_of_water,
! vanthoff, -b_cond in cloud_parameters,
! dopts_num increased to 29, particle attributes speed_x|y|z_sgs renamed
! rvar1|2|3
! wang_collision_kernel and turbulence_effects_on_collision replaced by
! collision_kernel, hall_kernel, palm_kernel, wang_kernel
!
! 809 2012-01-30 13:32:58Z marongas
! Bugfix: replaced .AND. and .NOT. with && and ! in the preprocessor directives
!
! 807 2012-01-25 11:53:51Z maronga
! New cpp directive "__check" implemented which is used by check_namelist_files.
! New parameter check_restart has been defined which is needed by 
! check_namelist_files only.
!
! 805 2012-01-17 15:53:28Z franke
! Bugfix collective_wait must be out of parallel branch for runs in serial mode
!
! 801 2012-01-10 17:30:36Z suehring
! Dimesion of sums_wsus_ws_l, ! sums_wsvs_ws_l, sums_us2_ws_l, sums_vs2_ws_l,
! sums_ws2_ws_l, sums_wspts_ws_l, sums_wsqs_ws_l, sums_wssas_ws_l increased.
! for thread-safe summation in advec_ws.
!
! RCS Log replace by Id keyword, revision history cleaned up
!
! Revision 1.95  2007/02/11 13:18:30  raasch
! version 3.1b (last under RCS control)
!
! Revision 1.1  1997/07/24 11:21:26  raasch
! Initial revision
!
!
!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of all variables
!>
!> @todo Add description for each variable
!------------------------------------------------------------------------------!


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of variables for special advection schemes.
!------------------------------------------------------------------------------!
 MODULE advection
 
    USE kinds

    REAL(wp), DIMENSION(:), ALLOCATABLE ::  aex  !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  bex  !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  dex  !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  eex  !< 
    
    SAVE

 END MODULE advection


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of all arrays defined on the computational grid.
!------------------------------------------------------------------------------!
 MODULE arrays_3d

    USE kinds

    REAL(wp), DIMENSION(:), ALLOCATABLE ::  c_u_m                  !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  c_u_m_l                !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  c_v_m                  !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  c_v_m_l                !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  c_w_m                  !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  c_w_m_l                !<    
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ddzu                   !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ddzu_pres              !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  dd2zu                  !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  dzu                    !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ddzw                   !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  dzw                    !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  hyp                    !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  inflow_damping_factor  !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  l_grid                 !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ptdf_x                 !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ptdf_y                 !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  p_surf                 !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  pt_surf                !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  pt_init                !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  qsws_surf              !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  q_init                 !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  q_surf                 !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  rdf                    !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  rdf_sc                 !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ref_state              !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  s_init                 !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  s_surf                 !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  sa_init                !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  shf_surf               !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  timenudge              !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  time_surf              !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  time_vert              !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  tmp_tnudge             !<
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ug                     !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  u_init                 !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  u_nzb_p1_for_vfc       !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  vg                     !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  v_init                 !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  v_nzb_p1_for_vfc       !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  w_subs                 !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  zu                     !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  zw                     !< 

    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  c_u                   !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  c_v                   !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  c_w                   !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  diss_s_e              !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  diss_s_nr             !<     
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  diss_s_pt             !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  diss_s_q              !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  diss_s_qr             !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  diss_s_s              !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  diss_s_sa             !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  diss_s_u              !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  diss_s_v              !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  diss_s_w              !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  dzu_mg                !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  dzw_mg                !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  flux_s_e              !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  flux_s_nr             !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  flux_s_pt             !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  flux_s_q              !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  flux_s_qr             !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  flux_s_s              !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  flux_s_sa             !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  flux_s_u              !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  flux_s_v              !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  flux_s_w              !<
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  f1_mg                 !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  f2_mg                 !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  f3_mg                 !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  mean_inflow_profiles  !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  precipitation_amount  !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  precipitation_rate    !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  ptnudge               !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  pt_slope_ref          !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  qnudge                !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  tnudge                !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  td_lsa_lpt            !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  td_lsa_q              !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  td_sub_lpt            !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  td_sub_q              !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  total_2d_a            !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  total_2d_o            !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  ug_vert               !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  unudge                !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  vnudge                !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  vg_vert               !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  wnudge                !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  wsubs_vert            !< 
    
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  d          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  de_dx      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  de_dy      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  de_dz      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  diss       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  diss_l_e   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  diss_l_nr  !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  diss_l_pt  !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  diss_l_q   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  diss_l_qr  !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  diss_l_s   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  diss_l_sa  !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  diss_l_u   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  diss_l_v   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  diss_l_w   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  flux_l_e   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  flux_l_nr  !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  flux_l_pt  !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  flux_l_q   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  flux_l_qr  !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  flux_l_s   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  flux_l_sa  !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  flux_l_u   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  flux_l_v   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  flux_l_w   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  kh         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  km         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  l_wall     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  prr        !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  p_loc      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  tend       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  tric       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  u_m_l      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  u_m_n      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  u_m_r      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  u_m_s      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  v_m_l      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  v_m_n      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  v_m_r      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  v_m_s      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  w_m_l      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  w_m_n      !<  
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  w_m_r      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  w_m_s      !< 

#if defined( __nopointer )
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  e          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  e_p        !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  nr         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  nr_p       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  p          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  prho       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  pt         !<
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  pt_p       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  q          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  q_p        !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  qc         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ql         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ql_c       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ql_v       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ql_vp      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  qr         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  qr_p       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  rho_ocean  !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  s          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  s_p        !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  sa         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  sa_p       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  te_m       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  tnr_m      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  tpt_m      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  tq_m       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  tqr_m      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ts_m       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  tsa_m      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  tu_m       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  tv_m       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  tw_m       !<  
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  u          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  u_p        !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  v          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  v_p        !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  vpt        !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  w          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  w_p        !< 
#else
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  e_1     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  e_2     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  e_3     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  p       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  prho_1  !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  nr_1    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  nr_2    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  nr_3    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  pt_1    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  pt_2    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  pt_3    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  q_1     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  q_2     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  q_3     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  qc_1    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ql_v    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ql_vp   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ql_1    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ql_2    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  qr_1    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  qr_2    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  qr_3    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  rho_1   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  s_1     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  s_2     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  s_3     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  sa_1    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  sa_2    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  sa_3    !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  u_1     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  u_2     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  u_3     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  v_1     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  v_2     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  v_3     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  vpt_1   !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  w_1     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  w_2     !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  w_3     !< 

    REAL(wp), DIMENSION(:,:,:), POINTER ::  e          !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  e_p        !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  nr         !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  nr_p       !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  prho       !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  pt         !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  pt_p       !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  q          !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  q_p        !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  qc         !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  ql         !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  ql_c       !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  qr         !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  qr_p       !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  rho_ocean  !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  s          !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  s_p        !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  sa         !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  sa_p       !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  te_m       !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  tnr_m      !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  tpt_m      !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  tq_m       !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  tqr_m      !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  ts_m       !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  tsa_m      !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  tu_m       !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  tv_m       !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  tw_m       !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  u          !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  u_p        !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  v          !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  v_p        !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  vpt        !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  w          !< 
    REAL(wp), DIMENSION(:,:,:), POINTER ::  w_p        !< 
#endif

    REAL(wp), DIMENSION(:,:,:,:), ALLOCATABLE ::  tri      !< 

    REAL(wp), DIMENSION(:), ALLOCATABLE ::  rho_air      !< air density profile on the uv grid
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  rho_air_zw   !< air density profile on the w grid
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  drho_air     !< inverse air density profile on the uv grid
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  drho_air_zw  !< inverse air density profile on the w grid

    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  rho_air_mg     !< air density profiles on the uv grid for multigrid
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  rho_air_zw_mg  !< air density profiles on the w grid for multigrid

    REAL(wp), DIMENSION(:), ALLOCATABLE ::  heatflux_input_conversion       !< conversion factor array for heatflux input
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  waterflux_input_conversion      !< conversion factor array for waterflux input
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  momentumflux_input_conversion   !< conversion factor array for momentumflux input
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  heatflux_output_conversion      !< conversion factor array for heatflux output
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  waterflux_output_conversion     !< conversion factor array for waterflux output
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  momentumflux_output_conversion  !< conversion factor array for momentumflux output

    SAVE

 END MODULE arrays_3d


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of variables needed for time-averaging of 2d/3d data.
!------------------------------------------------------------------------------!
 MODULE averaging
 
    USE kinds

    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  lwp_av                 !< Avg. liquid water path
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  precipitation_rate_av  !< Avg. of precipitation rate
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  ol_av                  !< Avg. of Obukhov length
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  qsws_av                !< Avg. of surface moisture flux
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  ssws_av                !< Avg. of surface scalar flux
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  shf_av                 !< Avg. of surface heat flux
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  ts_av                  !< Avg. of characteristic temperature scale
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  us_av                  !< Avg. of friction velocity
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  z0_av                  !< Avg. of roughness length for momentum
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  z0h_av                 !< Avg. of roughness length for heat
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  z0q_av                 !< Avg. of roughness length for moisture

    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  e_av          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  lpt_av        !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  nr_av         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  p_av          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  pc_av         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  pr_av         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  prr_av        !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  pt_av         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  q_av          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  qc_av         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ql_av         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ql_c_av       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ql_v_av       !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  ql_vp_av      !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  qr_av         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  qv_av         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  rho_ocean_av  !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  s_av          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  sa_av         !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  u_av          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  v_av          !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  vpt_av        !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE, TARGET ::  w_av          !< 
  
 END MODULE averaging


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of variables and constants for cloud physics.
!------------------------------------------------------------------------------!
 MODULE cloud_parameters
 
    USE kinds

    REAL(wp) ::  cp = 1005.0_wp         !< heat capacity of dry air (J kg-1 K-1)
    REAL(wp) ::  l_v = 2.5E+06_wp       !< latent heat of vaporization (J kg-1)
    REAL(wp) ::  l_d_cp                 !< l_v / cp
    REAL(wp) ::  l_d_r                  !< l_v / r_d
    REAL(wp) ::  l_d_rv                 !< l_v / r_v    
    REAL(wp) ::  rho_l = 1.0E3_wp       !< density of water (kg m-3)
    REAL(wp) ::  r_d = 287.0_wp         !< sp. gas const. dry air (J kg-1 K-1)
    REAL(wp) ::  r_v = 461.51_wp        !< sp. gas const. water vapor (J kg-1 K-1)


    REAL(wp), DIMENSION(:), ALLOCATABLE ::  hyrho   !<
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  pt_d_t  !<
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  t_d_pt  !<

    SAVE

 END MODULE cloud_parameters


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of general constants.
!------------------------------------------------------------------------------!
 MODULE constants
 
    USE kinds

    REAL(wp) ::  pi = 3.141592654_wp  !< 
    REAL(wp) ::  adv_mom_1            !< 
    REAL(wp) ::  adv_mom_3            !< 
    REAL(wp) ::  adv_mom_5            !< 
    REAL(wp) ::  adv_sca_1            !< 
    REAL(wp) ::  adv_sca_3            !< 
    REAL(wp) ::  adv_sca_5            !< 
    
    SAVE

 END MODULE constants

 
!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of parameters for program control
!------------------------------------------------------------------------------!
 MODULE control_parameters

    USE kinds

    TYPE plot_precision
       CHARACTER (LEN=8) ::  variable   !< 
       INTEGER(iwp)      ::  precision  !< 
    END TYPE plot_precision

    TYPE(plot_precision), DIMENSION(100) ::  plot_3d_precision =               &  !< 
                        (/ plot_precision( 'u', 2 ), plot_precision( 'v', 2 ), & 
                           plot_precision( 'w', 2 ), plot_precision( 'p', 5 ), & 
                           plot_precision( 'pt', 2 ),                          & 
                           ( plot_precision( ' ', 1 ), i9 = 1,95 ) /)

    TYPE file_status
       LOGICAL ::  opened         !< 
       LOGICAL ::  opened_before  !< 
    END TYPE file_status
    
    INTEGER, PARAMETER      ::  mask_xyz_dimension = 100  !< 
    INTEGER, PARAMETER      ::  max_masks = 50            !< 
    INTEGER(iwp), PARAMETER ::  varnamelength = 30        !< length of output variable names

    TYPE(file_status), DIMENSION(200+2*max_masks) ::                &  !< 
                             openfile = file_status(.FALSE.,.FALSE.)

    CHARACTER (LEN=1)    ::  cycle_mg = 'w'                               !< 
    CHARACTER (LEN=1)    ::  timestep_reason = ' '                        !< 
    CHARACTER (LEN=3)    ::  coupling_char = ''                           !< 
    CHARACTER (LEN=5)    ::  write_binary = 'false'                       !< 
    CHARACTER (LEN=8)    ::  most_method = 'newton'                       !< NAMELIST parameter defining method to be used to calculate Okukhov length, 
    CHARACTER (LEN=8)    ::  run_date                                     !< 
    CHARACTER (LEN=8)    ::  run_time                                     !< 
    CHARACTER (LEN=9)    ::  simulated_time_chr                           !< 
    CHARACTER (LEN=11)   ::  topography_grid_convention = ' '             !< 
    CHARACTER (LEN=12)   ::  version = ' '                                !< 
    CHARACTER (LEN=12)   ::  revision = ' '                               !< 
    CHARACTER (LEN=12)   ::  user_interface_current_revision = ' '        !< 
    CHARACTER (LEN=12)   ::  user_interface_required_revision = ' '       !< 
    CHARACTER (LEN=16)   ::  conserve_volume_flow_mode = 'default'        !< 
    CHARACTER (LEN=16)   ::  loop_optimization = 'default'                !< 
    CHARACTER (LEN=16)   ::  momentum_advec = 'ws-scheme'                 !< 
    CHARACTER (LEN=16)   ::  psolver = 'poisfft'                          !<  
    CHARACTER (LEN=16)   ::  scalar_advec = 'ws-scheme'                   !< 
    CHARACTER (LEN=20)   ::  approximation = 'boussinesq'                 !<  
    CHARACTER (LEN=40)   ::  flux_input_mode = 'approximation-specific'   !< 
    CHARACTER (LEN=40)   ::  flux_output_mode = 'approximation-specific'  !< 
    CHARACTER (LEN=20)   ::  bc_e_b = 'neumann'                           !< 
    CHARACTER (LEN=20)   ::  bc_lr = 'cyclic'                             !< 
    CHARACTER (LEN=20)   ::  bc_ns = 'cyclic'                             !< 
    CHARACTER (LEN=20)   ::  bc_p_b = 'neumann'                           !< 
    CHARACTER (LEN=20)   ::  bc_p_t = 'dirichlet'                         !< 
    CHARACTER (LEN=20)   ::  bc_pt_b = 'dirichlet'                        !< 
    CHARACTER (LEN=20)   ::  bc_pt_t = 'initial_gradient'                 !< 
    CHARACTER (LEN=20)   ::  bc_q_b = 'dirichlet'                         !< 
    CHARACTER (LEN=20)   ::  bc_q_t = 'neumann'                           !< 
    CHARACTER (LEN=20)   ::  bc_s_b = 'dirichlet'                         !< 
    CHARACTER (LEN=20)   ::  bc_s_t = 'initial_gradient'                  !< 
    CHARACTER (LEN=20)   ::  bc_sa_t = 'neumann'                          !< 
    CHARACTER (LEN=20)   ::  bc_uv_b = 'dirichlet'                        !< 
    CHARACTER (LEN=20)   ::  bc_uv_t = 'dirichlet'                        !< 
    CHARACTER (LEN=20)   ::  cloud_scheme = 'saturation_adjust'           !< 
    CHARACTER (LEN=20)   ::  coupling_mode = 'uncoupled'                  !<    
    CHARACTER (LEN=20)   ::  coupling_mode_remote = 'uncoupled'           !< 
    CHARACTER (LEN=20)   ::  dissipation_1d = 'detering'                  !< 
    CHARACTER (LEN=20)   ::  fft_method = 'system-specific'               !< 
    CHARACTER (LEN=20)   ::  mixing_length_1d = 'blackadar'               !< 
    CHARACTER (LEN=20)   ::  random_generator = 'random-parallel'         !< 
    CHARACTER (LEN=20)   ::  reference_state = 'initial_profile'          !<   
    CHARACTER (LEN=20)   ::  return_addres                                !< 
    CHARACTER (LEN=20)   ::  return_username                              !< 
    CHARACTER (LEN=20)   ::  timestep_scheme = 'runge-kutta-3'            !<        
    CHARACTER (LEN=40)   ::  avs_data_file                                !< 
    CHARACTER (LEN=40)   ::  topography = 'flat'                          !<  
    CHARACTER (LEN=64)   ::  host = ' '                                   !< 
    CHARACTER (LEN=80)   ::  log_message                                  !< 
    CHARACTER (LEN=80)   ::  run_identifier                               !<  
    CHARACTER (LEN=100)  ::  initializing_actions = ' '                   !< 
    CHARACTER (LEN=210)  ::  run_description_header                       !< 
    CHARACTER (LEN=1000) ::  message_string = ' '                         !< 

    CHARACTER (LEN=varnamelength), DIMENSION(500) ::  data_output = ' '       !< 
    CHARACTER (LEN=varnamelength), DIMENSION(500) ::  data_output_user = ' '  !< 
    CHARACTER (LEN=varnamelength), DIMENSION(500) ::  doav = ' '              !<
                                            
    CHARACTER (LEN=varnamelength), DIMENSION(max_masks,100) ::  data_output_masks = ' '       !< 
    CHARACTER (LEN=varnamelength), DIMENSION(max_masks,100) ::  data_output_masks_user = ' '  !< 

    CHARACTER (LEN=varnamelength), DIMENSION(300) ::  data_output_pr = ' '  !< 
    
    CHARACTER (LEN=varnamelength), DIMENSION(200) ::  data_output_pr_user = ' '  !< 
    
    CHARACTER (LEN=varnamelength), DIMENSION(max_masks,0:1,100) ::  domask = ' ' !< 
    
    CHARACTER (LEN=varnamelength), DIMENSION(0:1,500) ::  do2d = ' '  !< 
    CHARACTER (LEN=varnamelength), DIMENSION(0:1,500) ::  do3d = ' '  !< 

    INTEGER(iwp), PARAMETER ::  fl_max = 100     !< 
    INTEGER(iwp), PARAMETER ::  var_fl_max = 20  !< 
    
    INTEGER(iwp) ::  abort_mode = 1                    !< 
    INTEGER(iwp) ::  average_count_pr = 0              !< 
    INTEGER(iwp) ::  average_count_3d = 0              !< 
    INTEGER(iwp) ::  current_timestep_number = 0       !< 
    INTEGER(iwp) ::  coupling_topology = 0             !< 
    INTEGER(iwp) ::  dist_range = 0                    !< 
    INTEGER(iwp) ::  disturbance_level_ind_b           !< 
    INTEGER(iwp) ::  disturbance_level_ind_t           !< 
    INTEGER(iwp) ::  doav_n = 0                        !< 
    INTEGER(iwp) ::  dopr_n = 0                        !< 
    INTEGER(iwp) ::  dopr_time_count = 0               !< 
    INTEGER(iwp) ::  dopts_time_count = 0              !< 
    INTEGER(iwp) ::  dots_time_count = 0               !< 
    INTEGER(iwp) ::  do2d_xy_n = 0                     !< 
    INTEGER(iwp) ::  do2d_xz_n = 0                     !< 
    INTEGER(iwp) ::  do2d_yz_n = 0                     !< 
    INTEGER(iwp) ::  do3d_avs_n = 0                    !< 
    INTEGER(iwp) ::  dp_level_ind_b = 0                !< 
    INTEGER(iwp) ::  dvrp_filecount = 0                !< 
    INTEGER(iwp) ::  dz_stretch_level_index            !< 
    INTEGER(iwp) ::  ensemble_member_nr = 0            !< 
    INTEGER(iwp) ::  gamma_mg                          !< 
    INTEGER(iwp) ::  gathered_size                     !< 
    INTEGER(iwp) ::  grid_level                        !< 
    INTEGER(iwp) ::  ibc_e_b                           !< 
    INTEGER(iwp) ::  ibc_p_b                           !< 
    INTEGER(iwp) ::  ibc_p_t                           !< 
    INTEGER(iwp) ::  ibc_pt_b                          !< 
    INTEGER(iwp) ::  ibc_pt_t                          !< 
    INTEGER(iwp) ::  ibc_q_b                           !< 
    INTEGER(iwp) ::  ibc_q_t                           !< 
    INTEGER(iwp) ::  ibc_s_b                           !< 
    INTEGER(iwp) ::  ibc_s_t                           !< 
    INTEGER(iwp) ::  ibc_sa_t                          !< 
    INTEGER(iwp) ::  ibc_uv_b                          !< 
    INTEGER(iwp) ::  ibc_uv_t                          !< 
    INTEGER(iwp) ::  inflow_disturbance_begin = -1     !< 
    INTEGER(iwp) ::  inflow_disturbance_end = -1       !< 
    INTEGER(iwp) ::  intermediate_timestep_count       !< 
    INTEGER(iwp) ::  intermediate_timestep_count_max   !< 
    INTEGER(iwp) ::  io_group = 0                      !< 
    INTEGER(iwp) ::  io_blocks = 1                     !< 
    INTEGER(iwp) ::  iran = -1234567                   !< 
    INTEGER(iwp) ::  lod = 1                           !< level of detail, topography input parameter 
    INTEGER(iwp) ::  masks = 0                         !< 
    INTEGER(iwp) ::  maximum_grid_level                !< 
    INTEGER(iwp) ::  maximum_parallel_io_streams = -1  !< 
    INTEGER(iwp) ::  max_pr_user = 0                   !< 
    INTEGER(iwp) ::  mgcycles = 0                      !< 
    INTEGER(iwp) ::  mg_cycles = -1                    !< 
    INTEGER(iwp) ::  mg_switch_to_pe0_level = -1       !< 
    INTEGER(iwp) ::  mid                               !< 
    INTEGER(iwp) ::  nlsf = 1000                       !< 
    INTEGER(iwp) ::  ntnudge = 1000                    !< 
    INTEGER(iwp) ::  ngsrb = 2                         !< 
    INTEGER(iwp) ::  nr_timesteps_this_run = 0         !< 
    INTEGER(iwp) ::  nsor = 20                         !< 
    INTEGER(iwp) ::  nsor_ini = 100                    !< 
    INTEGER(iwp) ::  n_sor                             !< 
    INTEGER(iwp) ::  normalizing_region = 0            !< 
    INTEGER(iwp) ::  num_leg=0                         !< 
    INTEGER(iwp) ::  num_var_fl                        !< 
    INTEGER(iwp) ::  num_var_fl_user=0                 !< 
    INTEGER(iwp) ::  nz_do3d = -9999                   !< 
    INTEGER(iwp) ::  prt_time_count = 0                !< 
    INTEGER(iwp) ::  recycling_plane                   !< 
    INTEGER(iwp) ::  runnr = 0                         !< 
    INTEGER(iwp) ::  skip_do_avs = 0                   !< 
    INTEGER(iwp) ::  subdomain_size                    !< 
    INTEGER(iwp) ::  terminate_coupled = 0             !< 
    INTEGER(iwp) ::  terminate_coupled_remote = 0      !< 
    INTEGER(iwp) ::  timestep_count = 0                !< 

    INTEGER(iwp) ::  dist_nxl(0:1)                               !< 
    INTEGER(iwp) ::  dist_nxr(0:1)                               !< 
    INTEGER(iwp) ::  dist_nyn(0:1)                               !< 
    INTEGER(iwp) ::  dist_nys(0:1)                               !< 
    INTEGER(iwp) ::  do2d_no(0:1) = 0                            !< 
    INTEGER(iwp) ::  do2d_xy_time_count(0:1)                     !< 
    INTEGER(iwp) ::  do2d_xz_time_count(0:1)                     !< 
    INTEGER(iwp) ::  do2d_yz_time_count(0:1)                     !< 
    INTEGER(iwp) ::  do3d_no(0:1) = 0                            !< 
    INTEGER(iwp) ::  do3d_time_count(0:1)                        !< 
    INTEGER(iwp) ::  domask_no(max_masks,0:1) = 0                !< 
    INTEGER(iwp) ::  domask_time_count(max_masks,0:1)            !< 
    INTEGER(iwp) ::  mask_size(max_masks,3) = -1                 !< 
    INTEGER(iwp) ::  mask_size_l(max_masks,3) = -1               !< 
    INTEGER(iwp) ::  mask_start_l(max_masks,3) = -1              !< 
    INTEGER(iwp) ::  pt_vertical_gradient_level_ind(10) = -9999  !< 
    INTEGER(iwp) ::  q_vertical_gradient_level_ind(10) = -9999   !< 
    INTEGER(iwp) ::  s_vertical_gradient_level_ind(10) = -9999   !<                
    INTEGER(iwp) ::  sa_vertical_gradient_level_ind(10) = -9999  !< 
    INTEGER(iwp) ::  section(100,3)                              !< 
    INTEGER(iwp) ::  section_xy(100) = -9999                     !< 
    INTEGER(iwp) ::  section_xz(100) = -9999                     !< 
    INTEGER(iwp) ::  section_yz(100) = -9999                     !< 
    INTEGER(iwp) ::  ug_vertical_gradient_level_ind(10) = -9999  !< 
    INTEGER(iwp) ::  vg_vertical_gradient_level_ind(10) = -9999  !< 
    INTEGER(iwp) ::  subs_vertical_gradient_level_i(10) = -9999  !< 


    INTEGER(iwp), DIMENSION(0:1) ::  ntdim_2d_xy  !< 
    INTEGER(iwp), DIMENSION(0:1) ::  ntdim_2d_xz  !< 
    INTEGER(iwp), DIMENSION(0:1) ::  ntdim_2d_yz  !< 
    INTEGER(iwp), DIMENSION(0:1) ::  ntdim_3d     !< 

    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  grid_level_count  !< 

    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  mask_i         !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  mask_j         !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  mask_k         !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  mask_i_global  !<   
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  mask_j_global  !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  mask_k_global  !< 

    LOGICAL ::  bc_lr_cyc =.TRUE.                            !< 
    LOGICAL ::  bc_lr_dirrad = .FALSE.                       !< 
    LOGICAL ::  bc_lr_raddir = .FALSE.                       !< 
    LOGICAL ::  bc_ns_cyc = .TRUE.                           !< 
    LOGICAL ::  bc_ns_dirrad = .FALSE.                       !< 
    LOGICAL ::  bc_ns_raddir = .FALSE.                       !< 
    LOGICAL ::  call_microphysics_at_all_substeps = .FALSE.  !< 
    LOGICAL ::  call_psolver_at_all_substeps = .TRUE.        !< 
    LOGICAL ::  cloud_droplets = .FALSE.                     !< 
    LOGICAL ::  cloud_physics = .FALSE.                      !< 
    LOGICAL ::  cloud_top_radiation = .FALSE.                !< 
    LOGICAL ::  conserve_volume_flow = .FALSE.               !< 
    LOGICAL ::  constant_diffusion = .FALSE.                 !< 
    LOGICAL ::  constant_flux_layer = .TRUE.                 !< 
    LOGICAL ::  constant_heatflux = .TRUE.                   !<  
    LOGICAL ::  constant_top_heatflux = .TRUE.               !< 
    LOGICAL ::  constant_top_momentumflux = .FALSE.          !< 
    LOGICAL ::  constant_top_salinityflux = .TRUE.           !< 
    LOGICAL ::  constant_top_scalarflux = .TRUE.             !< 
    LOGICAL ::  constant_scalarflux = .TRUE.                 !< 
    LOGICAL ::  constant_waterflux = .TRUE.                  !< 
    LOGICAL ::  create_disturbances = .TRUE.                 !< 
    LOGICAL ::  data_output_2d_on_each_pe = .TRUE.           !< 
    LOGICAL ::  dissipation_control = .FALSE.                !< 
    LOGICAL ::  disturbance_created = .FALSE.                !< 
    LOGICAL ::  do2d_at_begin = .FALSE.                      !< 
    LOGICAL ::  do3d_at_begin = .FALSE.                      !< 
    LOGICAL ::  do_sum = .FALSE.                             !< 
    LOGICAL ::  dp_external = .FALSE.                        !< 
    LOGICAL ::  dp_smooth = .FALSE.                          !< 
    LOGICAL ::  dt_fixed = .FALSE.                           !< 
    LOGICAL ::  dt_3d_reached                                !< 
    LOGICAL ::  dt_3d_reached_l                              !< 
    LOGICAL ::  exchange_mg = .FALSE.                        !< 
    LOGICAL ::  first_call_lpm = .TRUE.                      !< 
    LOGICAL ::  force_print_header = .FALSE.                 !< 
    LOGICAL ::  galilei_transformation = .FALSE.             !< 
    LOGICAL ::  humidity = .FALSE.                           !< 
    LOGICAL ::  humidity_remote = .FALSE.                    !< 
    LOGICAL ::  inflow_l = .FALSE.                           !< 
    LOGICAL ::  inflow_n = .FALSE.                           !< 
    LOGICAL ::  inflow_r = .FALSE.                           !< 
    LOGICAL ::  inflow_s = .FALSE.                           !< 
    LOGICAL ::  large_scale_forcing = .FALSE.                !< 
    LOGICAL ::  large_scale_subsidence = .FALSE.             !< 
    LOGICAL ::  land_surface = .FALSE.                       !< flag parameter indicating wheather the lsm is used
    LOGICAL ::  lsf_exception = .FALSE.                      !< temporary flag for use of lsf with buildings on flat terrain
    LOGICAL ::  lsf_surf = .TRUE.                            !< 
    LOGICAL ::  lsf_vert = .TRUE.                            !< 
    LOGICAL ::  lptnudge = .FALSE.                           !< 
    LOGICAL ::  lqnudge = .FALSE.                            !< 
    LOGICAL ::  lunudge = .FALSE.                            !< 
    LOGICAL ::  lvnudge = .FALSE.                            !< 
    LOGICAL ::  lwnudge = .FALSE.                            !< 
    LOGICAL ::  masking_method = .FALSE.                     !< 
    LOGICAL ::  microphysics_sat_adjust = .FALSE.            !< 
    LOGICAL ::  microphysics_kessler = .FALSE.               !< 
    LOGICAL ::  microphysics_seifert = .FALSE.               !< 
    LOGICAL ::  mg_switch_to_pe0 = .FALSE.                   !< 
    LOGICAL ::  nest_bound_l = .FALSE.                       !< nested boundary on left side
    LOGICAL ::  nest_bound_n = .FALSE.                       !< nested boundary on north side
    LOGICAL ::  nest_bound_r = .FALSE.                       !< nested boundary on right side
    LOGICAL ::  nest_bound_s = .FALSE.                       !< nested boundary on south side
    LOGICAL ::  nest_domain  = .FALSE.                       !< domain is nested into a parent domain
    LOGICAL ::  neutral = .FALSE.                            !< 
    LOGICAL ::  nudging = .FALSE.                            !< 
    LOGICAL ::  ocean = .FALSE.                              !< 
    LOGICAL ::  outflow_l = .FALSE.                          !< 
    LOGICAL ::  outflow_n = .FALSE.                          !< 
    LOGICAL ::  outflow_r = .FALSE.                          !< 
    LOGICAL ::  outflow_s = .FALSE.                          !< 
    LOGICAL ::  passive_scalar = .FALSE.                     !< 
    LOGICAL ::  precipitation = .FALSE.                      !< 
    LOGICAL ::  random_heatflux = .FALSE.                    !< 
    LOGICAL ::  recycling_yshift = .FALSE.                   !< 
    LOGICAL ::  run_control_header = .FALSE.                 !< 
    LOGICAL ::  run_coupled = .TRUE.                         !< 
    LOGICAL ::  scalar_rayleigh_damping = .TRUE.             !< 
    LOGICAL ::  sloping_surface = .FALSE.                    !< 
    LOGICAL ::  stop_dt = .FALSE.                            !< 
    LOGICAL ::  synchronous_exchange = .FALSE.               !< 
    LOGICAL ::  terminate_run = .FALSE.                      !< 
    LOGICAL ::  transpose_compute_overlap = .FALSE.          !< 
    LOGICAL ::  turbulent_inflow = .FALSE.                   !< 
    LOGICAL ::  turbulent_outflow = .FALSE.                  !< flag for turbulent outflow condition
    LOGICAL ::  urban_surface = .FALSE.                      !< flag for urban surface model
    LOGICAL ::  use_cmax = .TRUE.                            !< 
    LOGICAL ::  use_initial_profile_as_reference = .FALSE.   !< 
    LOGICAL ::  use_prescribed_profile_data = .FALSE.        !< 
    LOGICAL ::  use_single_reference_value = .FALSE.         !< 
    LOGICAL ::  use_subsidence_tendencies = .FALSE.          !< 
    LOGICAL ::  use_surface_fluxes = .FALSE.                 !< 
    LOGICAL ::  use_top_fluxes = .FALSE.                     !< 
    LOGICAL ::  use_ug_for_galilei_tr = .TRUE.               !< 
    LOGICAL ::  use_upstream_for_tke = .FALSE.               !< 
    LOGICAL ::  virtual_flight = .FALSE.                     !< flag for virtual flight model
    LOGICAL ::  wall_adjustment = .TRUE.                     !< 
    LOGICAL ::  ws_scheme_sca = .FALSE.                      !< 
    LOGICAL ::  ws_scheme_mom = .FALSE.                      !< 
    LOGICAL ::  data_output_xy(0:1) = .FALSE.                !< 
    LOGICAL ::  data_output_xz(0:1) = .FALSE.                !< 
    LOGICAL ::  data_output_yz(0:1) = .FALSE.                !< 

    REAL(wp) ::  advected_distance_x = 0.0_wp                  !< 
    REAL(wp) ::  advected_distance_y = 0.0_wp                  !< 
    REAL(wp) ::  alpha_surface = 0.0_wp                        !< 
    REAL(wp) ::  atmos_ocean_sign = 1.0_wp                     !< 
    REAL(wp) ::  averaging_interval = 0.0_wp                   !< 
    REAL(wp) ::  averaging_interval_pr = 9999999.9_wp          !< 
    REAL(wp) ::  bc_pt_t_val                                   !< 
    REAL(wp) ::  bc_q_t_val                                    !< 
    REAL(wp) ::  bc_s_t_val                                    !< 
    REAL(wp) ::  bottom_salinityflux = 0.0_wp                  !< 
    REAL(wp) ::  building_height = 50.0_wp                     !< 
    REAL(wp) ::  building_length_x = 50.0_wp                   !< 
    REAL(wp) ::  building_length_y = 50.0_wp                   !< 
    REAL(wp) ::  building_wall_left = 9999999.9_wp             !< 
    REAL(wp) ::  building_wall_south = 9999999.9_wp            !< 
    REAL(wp) ::  canyon_height = 50.0_wp                       !< 
    REAL(wp) ::  canyon_width_x = 9999999.9_wp                 !< 
    REAL(wp) ::  canyon_width_y = 9999999.9_wp                 !< 
    REAL(wp) ::  canyon_wall_left = 9999999.9_wp               !< 
    REAL(wp) ::  canyon_wall_south = 9999999.9_wp              !< 
    REAL(wp) ::  cfl_factor = -1.0_wp                          !< 
    REAL(wp) ::  cos_alpha_surface                             !< 
    REAL(wp) ::  coupling_start_time = 0.0_wp                  !< 
    REAL(wp) ::  disturbance_amplitude = 0.25_wp               !< 
    REAL(wp) ::  disturbance_energy_limit = 0.01_wp            !< 
    REAL(wp) ::  disturbance_level_b = -9999999.9_wp           !< 
    REAL(wp) ::  disturbance_level_t = -9999999.9_wp           !< 
    REAL(wp) ::  dp_level_b = 0.0_wp                           !< 
    REAL(wp) ::  dt = -1.0_wp                                  !< 
    REAL(wp) ::  dt_averaging_input = 0.0_wp                   !< 
    REAL(wp) ::  dt_averaging_input_pr = 9999999.9_wp          !< 
    REAL(wp) ::  dt_coupling = 9999999.9_wp                    !< 
    REAL(wp) ::  dt_data_output = 9999999.9_wp                 !< 
    REAL(wp) ::  dt_data_output_av = 9999999.9_wp              !< 
    REAL(wp) ::  dt_disturb = 9999999.9_wp                     !< 
    REAL(wp) ::  dt_dopr = 9999999.9_wp                        !< 
    REAL(wp) ::  dt_dopr_listing = 9999999.9_wp                !< 
    REAL(wp) ::  dt_dopts = 9999999.9_wp                       !< 
    REAL(wp) ::  dt_dots = 9999999.9_wp                        !< 
    REAL(wp) ::  dt_do2d_xy = 9999999.9_wp                     !< 
    REAL(wp) ::  dt_do2d_xz = 9999999.9_wp                     !< 
    REAL(wp) ::  dt_do2d_yz = 9999999.9_wp                     !< 
    REAL(wp) ::  dt_do3d = 9999999.9_wp                        !< 
    REAL(wp) ::  dt_dvrp = 9999999.9_wp                        !< 
    REAL(wp) ::  dt_max = 20.0_wp                              !< 
    REAL(wp) ::  dt_restart = 9999999.9_wp                     !< 
    REAL(wp) ::  dt_run_control = 60.0_wp                      !< 
    REAL(wp) ::  dt_3d = 1.0_wp                                !< 
    REAL(wp) ::  dz = -1.0_wp                                  !< 
    REAL(wp) ::  dz_max = 9999999.9_wp                         !< 
    REAL(wp) ::  dz_stretch_factor = 1.08_wp                   !< 
    REAL(wp) ::  dz_stretch_level = 100000.0_wp                !< 
    REAL(wp) ::  e_init = 0.0_wp                               !< 
    REAL(wp) ::  e_min = 0.0_wp                                !< 
    REAL(wp) ::  end_time = 0.0_wp                             !< 
    REAL(wp) ::  f = 0.0_wp                                    !< 
    REAL(wp) ::  fs = 0.0_wp                                   !< 
    REAL(wp) ::  g = 9.81_wp                                   !< 
    REAL(wp) ::  inflow_damping_height = 9999999.9_wp          !< 
    REAL(wp) ::  inflow_damping_width = 9999999.9_wp           !< 
    REAL(wp) ::  kappa = 0.4_wp                                !< 
    REAL(wp) ::  km_constant = -1.0_wp                         !< 
    REAL(wp) ::  mask_scale_x = 1.0_wp                         !< 
    REAL(wp) ::  mask_scale_y = 1.0_wp                         !< 
    REAL(wp) ::  mask_scale_z = 1.0_wp                         !< 
    REAL(wp) ::  maximum_cpu_time_allowed = 0.0_wp             !< 
    REAL(wp) ::  molecular_viscosity = 1.461E-5_wp             !< 
    REAL(wp) ::  old_dt = 1.0E-10_wp                           !< 
    REAL(wp) ::  omega = 7.29212E-5_wp                         !< 
    REAL(wp) ::  omega_sor = 1.8_wp                            !< 
    REAL(wp) ::  outflow_source_plane = -9999999.9_wp          !< x-position of outflow-source plane (turbulent outflow method)
    REAL(wp) ::  particle_maximum_age = 9999999.9_wp           !< 
    REAL(wp) ::  phi = 55.0_wp                                 !< 
    REAL(wp) ::  prandtl_number = 1.0_wp                       !< 
    REAL(wp) ::  precipitation_amount_interval = 9999999.9_wp  !< 
    REAL(wp) ::  prho_reference                                !< 
    REAL(wp) ::  pt_damping_factor = 0.0_wp                    !< 
    REAL(wp) ::  pt_damping_width = 0.0_wp                     !< 
    REAL(wp) ::  pt_reference = 9999999.9_wp                   !< 
    REAL(wp) ::  pt_slope_offset = 0.0_wp                      !< 
    REAL(wp) ::  pt_surface = 300.0_wp                         !< 
    REAL(wp) ::  pt_surface_initial_change = 0.0_wp            !< 
    REAL(wp) ::  q_surface = 0.0_wp                            !< 
    REAL(wp) ::  q_surface_initial_change = 0.0_wp             !< 
    REAL(wp) ::  rayleigh_damping_factor = -1.0_wp             !< 
    REAL(wp) ::  rayleigh_damping_height = -1.0_wp             !< 
    REAL(wp) ::  recycling_width = 9999999.9_wp                !< 
    REAL(wp) ::  residual_limit = 1.0E-4_wp                    !< 
    REAL(wp) ::  restart_time = 9999999.9_wp                   !< 
    REAL(wp) ::  rho_reference                                 !< 
    REAL(wp) ::  rho_surface                                   !< 
    REAL(wp) ::  roughness_length = 0.1_wp                     !< 
    REAL(wp) ::  sa_surface = 35.0_wp                          !< 
    REAL(wp) ::  simulated_time = 0.0_wp                       !< 
    REAL(wp) ::  simulated_time_at_begin                       !< 
    REAL(wp) ::  sin_alpha_surface                             !< 
    REAL(wp) ::  skip_time_data_output = 0.0_wp                !< 
    REAL(wp) ::  skip_time_data_output_av = 9999999.9_wp       !< 
    REAL(wp) ::  skip_time_dopr = 9999999.9_wp                 !< 
    REAL(wp) ::  skip_time_do2d_xy = 9999999.9_wp              !< 
    REAL(wp) ::  skip_time_do2d_xz = 9999999.9_wp              !< 
    REAL(wp) ::  skip_time_do2d_yz = 9999999.9_wp              !< 
    REAL(wp) ::  skip_time_do3d = 9999999.9_wp                 !< 
    REAL(wp) ::  surface_heatflux = 9999999.9_wp               !< 
    REAL(wp) ::  surface_pressure = 1013.25_wp                 !< 
    REAL(wp) ::  surface_scalarflux = 9999999.9_wp             !< 
    REAL(wp) ::  surface_waterflux = 9999999.9_wp              !< 
    REAL(wp) ::  s_surface = 0.0_wp                            !< 
    REAL(wp) ::  s_surface_initial_change = 0.0_wp             !< 
    REAL(wp) ::  termination_time_needed = -1.0_wp             !< 
    REAL(wp) ::  time_coupling = 0.0_wp                        !< 
    REAL(wp) ::  time_disturb = 0.0_wp                         !< 
    REAL(wp) ::  time_dopr = 0.0_wp                            !< 
    REAL(wp) ::  time_dopr_av = 0.0_wp                         !< 
    REAL(wp) ::  time_dopr_listing = 0.0_wp                    !< 
    REAL(wp) ::  time_dopts = 0.0_wp                           !< 
    REAL(wp) ::  time_dosp = 0.0_wp                            !< 
    REAL(wp) ::  time_dosp_av = 0.0_wp                         !< 
    REAL(wp) ::  time_dots = 0.0_wp                            !< 
    REAL(wp) ::  time_do2d_xy = 0.0_wp                         !< 
    REAL(wp) ::  time_do2d_xz = 0.0_wp                         !< 
    REAL(wp) ::  time_do2d_yz = 0.0_wp                         !< 
    REAL(wp) ::  time_do3d = 0.0_wp                            !< 
    REAL(wp) ::  time_do_av = 0.0_wp                           !< 
    REAL(wp) ::  time_do_sla = 0.0_wp                          !< 
    REAL(wp) ::  time_dvrp = 0.0_wp                            !< 
    REAL(wp) ::  time_restart = 9999999.9_wp                   !< 
    REAL(wp) ::  time_run_control = 0.0_wp                     !< 
    REAL(wp) ::  time_since_reference_point                    !< 
    REAL(wp) ::  top_heatflux = 9999999.9_wp                   !< 
    REAL(wp) ::  top_momentumflux_u = 9999999.9_wp             !< 
    REAL(wp) ::  top_momentumflux_v = 9999999.9_wp             !< 
    REAL(wp) ::  top_salinityflux = 9999999.9_wp               !< 
    REAL(wp) ::  top_scalarflux = 9999999.9_wp                 !< 
    REAL(wp) ::  tunnel_height = 9999999.9_wp                  !< height of tunnel outer wall
    REAL(wp) ::  tunnel_length = 9999999.9_wp                  !< tunnel length
    REAL(wp) ::  tunnel_width_x = 9999999.9_wp                 !< tunnel width in x, with respect to outer wall
    REAL(wp) ::  tunnel_width_y = 9999999.9_wp                 !< tunnel width in y, with respect to outer wall
    REAL(wp) ::  tunnel_wall_depth = 9999999.9_wp              !< tunnel wall depth
    REAL(wp) ::  ug_surface = 0.0_wp                           !< 
    REAL(wp) ::  u_bulk = 0.0_wp                               !< 
    REAL(wp) ::  u_gtrans = 0.0_wp                             !< 
    REAL(wp) ::  vg_surface = 0.0_wp                           !< 
    REAL(wp) ::  vpt_reference = 9999999.9_wp                  !< 
    REAL(wp) ::  v_bulk = 0.0_wp                               !< 
    REAL(wp) ::  v_gtrans = 0.0_wp                             !< 
    REAL(wp) ::  wall_adjustment_factor = 1.8_wp               !< 
    REAL(wp) ::  zeta_max = 20.0_wp                            !< Maximum value of zeta = z/L
    REAL(wp) ::  zeta_min = -9990.0_wp                         !< Minimum value of zeta = z/L
    REAL(wp) ::  z_max_do2d = -1.0_wp                          !< 
    REAL(wp) ::  z0h_factor = 1.0_wp                           !< 

    REAL(wp) ::  do2d_xy_last_time(0:1) = -1.0_wp                  !< 
    REAL(wp) ::  do2d_xz_last_time(0:1) = -1.0_wp                  !< 
    REAL(wp) ::  do2d_yz_last_time(0:1) = -1.0_wp                  !< 
    REAL(wp) ::  dpdxy(1:2) = 0.0_wp                               !< 
    REAL(wp) ::  dt_domask(max_masks) = 9999999.9_wp               !< 
    REAL(wp) ::  mask_scale(3)                                     !< 
    REAL(wp) ::  pt_vertical_gradient(10) = 0.0_wp                 !< 
    REAL(wp) ::  pt_vertical_gradient_level(10) = -1.0_wp          !< 
    REAL(wp) ::  q_vertical_gradient(10) = 0.0_wp                  !< 
    REAL(wp) ::  q_vertical_gradient_level(10) = -1.0_wp           !< 
    REAL(wp) ::  s_vertical_gradient(10) = 0.0_wp                  !< 
    REAL(wp) ::  s_vertical_gradient_level(10) = -1.0_wp           !< 
    REAL(wp) ::  sa_vertical_gradient(10) = 0.0_wp                 !< 
    REAL(wp) ::  sa_vertical_gradient_level(10) = -1.0_wp          !< 
    REAL(wp) ::  skip_time_domask(max_masks) = 9999999.9_wp        !< 
    REAL(wp) ::  threshold(20) = 0.0_wp                            !< 
    REAL(wp) ::  time_domask(max_masks) = 0.0_wp                   !< 
    REAL(wp) ::  tsc(10) = (/ 1.0_wp, 1.0_wp, 0.0_wp, 0.0_wp, &    !< 
                 0.0_wp, 0.0_wp, 0.0_wp, 0.0_wp, 0.0_wp, 0.0_wp /)
    REAL(wp) ::  u_profile(100) = 9999999.9_wp                     !< 
    REAL(wp) ::  uv_heights(100) = 9999999.9_wp                    !< 
    REAL(wp) ::  v_profile(100) = 9999999.9_wp                     !< 
    REAL(wp) ::  ug_vertical_gradient(10) = 0.0_wp                 !< 
    REAL(wp) ::  ug_vertical_gradient_level(10) = -9999999.9_wp    !< 
    REAL(wp) ::  vg_vertical_gradient(10) = 0.0_wp                 !< 
    REAL(wp) ::  vg_vertical_gradient_level(10) = -9999999.9_wp    !< 
    REAL(wp) ::  volume_flow(1:3) = 0.0_wp                         !< 
    REAL(wp) ::  volume_flow_area(1:3) = 0.0_wp                    !< 
    REAL(wp) ::  volume_flow_initial(1:3) = 0.0_wp                 !< 
    REAL(wp) ::  wall_heatflux(0:4) = 0.0_wp                       !< 
    REAL(wp) ::  wall_humidityflux(0:4) = 0.0_wp                   !< 
    REAL(wp) ::  wall_salinityflux(0:4) = 0.0_wp                   !< 
    REAL(wp) ::  wall_scalarflux(0:4) = 0.0_wp                     !< 
    REAL(wp) ::  subs_vertical_gradient(10) = 0.0_wp               !< 
    REAL(wp) ::  subs_vertical_gradient_level(10) = -9999999.9_wp  !< 

    REAL(wp), DIMENSION(:), ALLOCATABLE ::  dp_smooth_factor  !< 

    REAL(wp), DIMENSION(max_masks,mask_xyz_dimension) ::  mask_x = -1.0_wp  !< 
    REAL(wp), DIMENSION(max_masks,mask_xyz_dimension) ::  mask_y = -1.0_wp  !< 
    REAL(wp), DIMENSION(max_masks,mask_xyz_dimension) ::  mask_z = -1.0_wp  !<
    
    REAL(wp), DIMENSION(max_masks,3) ::  mask_x_loop = -1.0_wp  !< 
    REAL(wp), DIMENSION(max_masks,3) ::  mask_y_loop = -1.0_wp  !< 
    REAL(wp), DIMENSION(max_masks,3) ::  mask_z_loop = -1.0_wp  !< 
    
!
!--    internal mask arrays ("mask,dimension,selection")
       REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  mask       !< 
       REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  mask_loop  !< 

    SAVE

 END MODULE control_parameters


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of variables used with dvrp-software.
!------------------------------------------------------------------------------!
 MODULE dvrp_variables

    USE kinds

    CHARACTER (LEN=10) ::  dvrp_output = 'rtsp'        !< 
    CHARACTER (LEN=10) ::  particle_color = 'none'     !< 
    CHARACTER (LEN=10) ::  particle_dvrpsize = 'none'  !< 

    CHARACTER (LEN=20), DIMENSION(10) ::  mode_dvrp = &  !< 
                                     (/ ( '                    ', i9 = 1,10 ) /)

    CHARACTER (LEN=80) ::  dvrp_directory = 'default'                     !< 
    CHARACTER (LEN=80) ::  dvrp_file      = 'default'                     !< 
    CHARACTER (LEN=80) ::  dvrp_host      = 'origin.rvs.uni-hannover.de'  !< 
    CHARACTER (LEN=80) ::  dvrp_password  = '********'                    !< 
    CHARACTER (LEN=80) ::  dvrp_username  = ' '                           !< 

    INTEGER(iwp) ::  cluster_size = 1                   !< 
    INTEGER(iwp) ::  dvrp_colortable_entries = 4        !< 
    INTEGER(iwp) ::  dvrp_colortable_entries_prt = 22   !< 
    INTEGER(iwp) ::  islice_dvrp                        !< 
    INTEGER(iwp) ::  nx_dvrp                            !< 
    INTEGER(iwp) ::  nxl_dvrp                           !< 
    INTEGER(iwp) ::  nxr_dvrp                           !< 
    INTEGER(iwp) ::  ny_dvrp                            !< 
    INTEGER(iwp) ::  nyn_dvrp                           !< 
    INTEGER(iwp) ::  nys_dvrp                           !< 
    INTEGER(iwp) ::  nz_dvrp, pathlines_fadeintime = 5  !< 
    INTEGER(iwp) ::  pathlines_fadeouttime = 5          !< 
    INTEGER(iwp) ::  pathlines_linecount = 1000         !< 
    INTEGER(iwp) ::  pathlines_maxhistory = 40          !< 
    INTEGER(iwp) ::  pathlines_wavecount = 10           !< 
    INTEGER(iwp) ::  pathlines_wavetime = 50            !< 
    INTEGER(iwp) ::  vc_gradient_normals = 0            !< 
    INTEGER(iwp) ::  vc_mode = 0                        !< 
    INTEGER(iwp) ::  vc_size_x = 2                      !< 
    INTEGER(iwp) ::  vc_size_y = 2                      !< 
    INTEGER(iwp) ::  vc_size_z = 2                      !< 

    INTEGER(iwp), DIMENSION(10) ::  slicer_position_dvrp  !< 

    LOGICAL ::  cyclic_dvrp = .FALSE.                      !< 
    LOGICAL ::  dvrp_overlap                               !< 
    LOGICAL ::  dvrp_total_overlap                         !< 
    LOGICAL ::  local_dvrserver_running                    !< 
    LOGICAL ::  lock_steering_update = .FALSE.             !< 
    LOGICAL ::  use_seperate_pe_for_dvrp_output = .FALSE.  !< 

    REAL(wp) ::  clip_dvrp_l = 9999999.9_wp  !< 
    REAL(wp) ::  clip_dvrp_n = 9999999.9_wp  !< 
    REAL(wp) ::  clip_dvrp_r = 9999999.9_wp  !< 
    REAL(wp) ::  clip_dvrp_s = 9999999.9_wp  !< 
    REAL(wp) ::  superelevation = 1.0_wp     !< 
    REAL(wp) ::  superelevation_x = 1.0_wp   !< 
    REAL(wp) ::  superelevation_y = 1.0_wp   !< 
    REAL(wp) ::  vc_alpha = 38.0_wp          !< 

    REAL(wp), DIMENSION(2) ::  color_interval = (/ 0.0_wp, 1.0_wp /)     !< 
    REAL(wp), DIMENSION(2) ::  dvrpsize_interval = (/ 0.0_wp, 1.0_wp /)  !< 

    REAL(wp), DIMENSION(3) ::  groundplate_color = (/ 0.0_wp, 0.6_wp, 0.0_wp /)  !< 
    REAL(wp), DIMENSION(3) ::  topography_color = (/ 0.8_wp, 0.7_wp, 0.6_wp /)   !< 

    REAL(wp), DIMENSION(2,10) ::  slicer_range_limits_dvrp  !< 

    REAL(wp), DIMENSION(3,10) ::  isosurface_color  !<

    REAL(sp), DIMENSION(2,100) ::  interval_values_dvrp          !< 
    REAL(sp), DIMENSION(2,100) ::  interval_values_dvrp_prt      !< 
    REAL(sp), DIMENSION(2,100) ::  interval_h_dvrp               !< 
    REAL(sp), DIMENSION(2,100) ::  interval_h_dvrp_prt           !< 
    REAL(sp), DIMENSION(2,100) ::  interval_l_dvrp = 0.5_sp      !< 
    REAL(sp), DIMENSION(2,100) ::  interval_l_dvrp_prt = 0.5_sp  !< 
    REAL(sp), DIMENSION(2,100) ::  interval_s_dvrp = 1.0_sp      !< 
    REAL(sp), DIMENSION(2,100) ::  interval_s_dvrp_prt = 1.0_sp  !< 
    REAL(sp), DIMENSION(2,100) ::  interval_a_dvrp = 0.0_sp      !< 
    REAL(sp), DIMENSION(2,100) ::  interval_a_dvrp_prt = 0.0_sp  !< 

    DATA  slicer_range_limits_dvrp / -1.0_wp, 1.0_wp, -1.0_wp, 1.0_wp, -1.0_wp, 1.0_wp, &  !< 
                                     -1.0_wp, 1.0_wp, -1.0_wp, 1.0_wp, -1.0_wp, 1.0_wp, &
                                     -1.0_wp, 1.0_wp, -1.0_wp, 1.0_wp, -1.0_wp, 1.0_wp, &
                                     -1.0_wp, 1.0_wp /

    DATA  isosurface_color / 0.9_wp, 0.9_wp, 0.9_wp,  0.8_wp, 0.1_wp, 0.1_wp,  0.1_wp, 0.1_wp, 0.8_wp, &  !< 
                             0.1_wp, 0.8_wp, 0.1_wp,  0.6_wp, 0.1_wp, 0.1_wp,  0.1_wp, 0.1_wp, 0.6_wp, &
                             0.1_wp, 0.6_wp, 0.1_wp,  0.4_wp, 0.1_wp, 0.1_wp,  0.1_wp, 0.1_wp, 0.4_wp, &
                             0.1_wp, 0.4_wp, 0.1_wp /

    DATA  interval_h_dvrp / 270.0_wp, 225.0_wp, 225.0_wp, 180.0_wp, 70.0_wp, 25.0_wp, &  !< 
                            25.0_wp, -25.0_wp, 192 * 0.0_wp /

    DATA  interval_h_dvrp_prt / 270.0_wp, 225.0_wp, 225.0_wp, 180.0_wp, 70.0_wp, 25.0_wp, &  !< 
                                25.0_wp, -25.0_wp, 192 * 0.0_wp /

    REAL(sp), DIMENSION(:), ALLOCATABLE ::  xcoor_dvrp  !< 
    REAL(sp), DIMENSION(:), ALLOCATABLE ::  ycoor_dvrp  !< 
    REAL(sp), DIMENSION(:), ALLOCATABLE ::  zcoor_dvrp  !< 

    TYPE steering
       CHARACTER (LEN=24) ::  name  !<
       REAL(sp)           ::  min   !<
       REAL(sp)           ::  max   !<
       INTEGER(iwp)       ::  imin  !<
       INTEGER(iwp)       ::  imax  !<
    END TYPE steering

    TYPE(steering), DIMENSION(:), ALLOCATABLE ::  steering_dvrp  !< 

    SAVE

 END MODULE dvrp_variables


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of grid spacings.
!------------------------------------------------------------------------------!
 MODULE grid_variables

    USE kinds

    REAL(wp) ::  ddx          !< 
    REAL(wp) ::  ddx2         !< 
    REAL(wp) ::  dx = 1.0_wp  !< 
    REAL(wp) ::  dx2          !< 
    REAL(wp) ::  ddy          !< 
    REAL(wp) ::  ddy2         !< 
    REAL(wp) ::  dy = 1.0_wp  !< 
    REAL(wp) ::  dy2          !< 

    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ddx2_mg  !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  ddy2_mg  !< 

    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  zu_s_inner  !<  
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  zw_w_inner  !< 
                                              
    SAVE

 END MODULE grid_variables


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of array bounds, number of gridpoints, and wall flag arrays.
!------------------------------------------------------------------------------!
 MODULE indices

    USE kinds

    INTEGER(iwp) ::  nbgp = 3     !< 
    INTEGER(iwp) ::  ngp_sums     !< 
    INTEGER(iwp) ::  ngp_sums_ls  !< 
    INTEGER(iwp) ::  nnx          !< 
    INTEGER(iwp) ::  nx = 0       !< 
    INTEGER(iwp) ::  nx_a         !< 
    INTEGER(iwp) ::  nx_o         !< 
    INTEGER(iwp) ::  nxl          !< 
    INTEGER(iwp) ::  nxlg         !< 
    INTEGER(iwp) ::  nxlu         !< 
    INTEGER(iwp) ::  nxr          !< 
    INTEGER(iwp) ::  nxrg         !< 
    INTEGER(iwp) ::  nx_on_file   !< 
    INTEGER(iwp) ::  nny          !< 
    INTEGER(iwp) ::  ny = 0       !< 
    INTEGER(iwp) ::  ny_a         !< 
    INTEGER(iwp) ::  ny_o         !< 
    INTEGER(iwp) ::  nyn          !< 
    INTEGER(iwp) ::  nyng         !< 
    INTEGER(iwp) ::  nys          !< 
    INTEGER(iwp) ::  nysg         !< 
    INTEGER(iwp) ::  nysv         !< 
    INTEGER(iwp) ::  ny_on_file   !< 
    INTEGER(iwp) ::  nnz          !< 
    INTEGER(iwp) ::  nz = 0       !< 
    INTEGER(iwp) ::  nzb          !< 
    INTEGER(iwp) ::  nzb_diff     !< 
    INTEGER(iwp) ::  nzb_max      !< 
    INTEGER(iwp) ::  nzt          !< 

    INTEGER(idp), DIMENSION(:), ALLOCATABLE ::  ngp_3d        !< 
    INTEGER(idp), DIMENSION(:), ALLOCATABLE ::  ngp_3d_inner  !< ! need to have 64 bit for grids > 2E9
                    
    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  ngp_2dh  !< 
    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  nxl_mg   !< 
    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  nxr_mg   !< 
    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  nyn_mg   !< 
    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  nys_mg   !< 
    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  nzt_mg   !< 


    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  ngp_2dh_outer     !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  ngp_2dh_s_inner   !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  mg_loc_ind        !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  nzb_diff_s_inner  !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  nzb_diff_s_outer  !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  nzb_inner         !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  nzb_outer         !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  nzb_s_inner       !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  nzb_s_outer       !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  nzb_u_inner       !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  nzb_u_outer       !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  nzb_v_inner       !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  nzb_v_outer       !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  nzb_w_inner       !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  nzb_w_outer       !< 

    INTEGER(iwp), DIMENSION(:,:,:), POINTER ::  flags  !<

    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE,  TARGET ::  wall_flags_1   !< 
    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE,  TARGET ::  wall_flags_2   !< 
    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE,  TARGET ::  wall_flags_3   !<
    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE,  TARGET ::  wall_flags_4   !< 
    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE,  TARGET ::  wall_flags_5   !< 
    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE,  TARGET ::  wall_flags_6   !< 
    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE,  TARGET ::  wall_flags_7   !< 
    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE,  TARGET ::  wall_flags_8   !< 
    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE,  TARGET ::  wall_flags_9   !< 
    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE,  TARGET ::  wall_flags_10  !< 

    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE ::  advc_flags_1            !< flags used to degrade order of advection scheme
    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE ::  advc_flags_2            !< flags used to degrade order of advection scheme
    INTEGER(iwp), DIMENSION(:,:,:), ALLOCATABLE ::  wall_flags_0            !< flags to mask topography

    SAVE

 END MODULE indices


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Interfaces for special subroutines which use optional parameters.
!------------------------------------------------------------------------------!
 MODULE interfaces

    INTERFACE

!------------------------------------------------------------------------------!
! Description:
! ------------
!> @todo Missing subroutine description.
!------------------------------------------------------------------------------!
       SUBROUTINE global_min_max ( i1, i2, j1, j2, k1, k2, feld, mode, offset, &
                                   wert, wert_ijk, wert1, wert1_ijk )

          USE kinds

          CHARACTER (LEN=*), INTENT(IN) ::  mode                     !< 
          INTEGER(iwp), INTENT(IN)      ::  i1                       !< 
          INTEGER(iwp), INTENT(IN)      ::  i2                       !< 
          INTEGER(iwp), INTENT(IN)      ::  j1                       !< 
          INTEGER(iwp), INTENT(IN)      ::  j2                       !< 
          INTEGER(iwp), INTENT(IN)      ::  k1                       !< 
          INTEGER(iwp), INTENT(IN)      ::  k2                       !<  
          INTEGER(iwp)                  ::  wert_ijk(3)              !< 
          INTEGER(iwp), OPTIONAL        ::  wert1_ijk(3)             !<  
          REAL(wp)                      ::  offset                   !< 
          REAL(wp)                      ::  wert                     !< 
          REAL(wp), OPTIONAL            ::  wert1                    !< 
          REAL(wp), INTENT(IN)          ::  feld(i1:i2,j1:j2,k1:k2)  !< 

       END SUBROUTINE global_min_max

    END INTERFACE

    SAVE

 END MODULE interfaces


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Interfaces for subroutines with pointer arguments called in
!> prognostic_equations.
!------------------------------------------------------------------------------!
 MODULE pointer_interfaces

    INTERFACE

!------------------------------------------------------------------------------!
! Description:
! ------------
!> @todo Missing subroutine description.
!------------------------------------------------------------------------------!
       SUBROUTINE advec_s_bc( sk, sk_char )

          USE kinds

          CHARACTER (LEN=*), INTENT(IN) ::  sk_char  !< 
#if defined( __nopointer )
          REAL(wp), DIMENSION(:,:,:) ::  sk  !< 
#else
          REAL(wp), DIMENSION(:,:,:), POINTER ::  sk  !< 
#endif
       END SUBROUTINE advec_s_bc

    END INTERFACE

    SAVE

 END MODULE pointer_interfaces


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of variables for the 1D-model.
!------------------------------------------------------------------------------!
 MODULE model_1d

    USE kinds

    INTEGER(iwp) ::  current_timestep_number_1d = 0  !< 
    INTEGER(iwp) ::  damp_level_ind_1d               !< 

    LOGICAL ::  run_control_header_1d = .FALSE.  !< 
    LOGICAL ::  stop_dt_1d = .FALSE.             !< 

    REAL(wp) ::  damp_level_1d = -1.0_wp       !< 
    REAL(wp) ::  dt_1d = 60.0_wp               !< 
    REAL(wp) ::  dt_max_1d = 300.0_wp          !< 
    REAL(wp) ::  dt_pr_1d = 9999999.9_wp       !< 
    REAL(wp) ::  dt_run_control_1d = 60.0_wp   !< 
    REAL(wp) ::  end_time_1d = 864000.0_wp     !< 
    REAL(wp) ::  old_dt_1d = 1.0E-10_wp        !< 
    REAL(wp) ::  qs1d                          !< 
    REAL(wp) ::  simulated_time_1d = 0.0_wp    !< 
    REAL(wp) ::  time_pr_1d = 0.0_wp           !< 
    REAL(wp) ::  time_run_control_1d = 0.0_wp  !< 
    REAL(wp) ::  ts1d                          !< 
    REAL(wp) ::  us1d                          !< 
    REAL(wp) ::  usws1d                        !< 
    REAL(wp) ::  vsws1d                        !<               
    REAL(wp) ::  z01d                          !< 
    REAL(wp) ::  z0h1d                         !< 


    REAL(wp), DIMENSION(:), ALLOCATABLE ::  e1d      !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  e1d_p    !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  kh1d     !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  km1d     !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  l_black  !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  l1d      !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  rif1d    !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  te_e     !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  te_em    !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  te_u     !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  te_um    !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  te_v     !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  te_vm    !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  u1d      !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  u1d_p    !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  v1d      !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  v1d_p    !< 

    SAVE

 END MODULE model_1d




!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of variables which define processor topology and the exchange of
!> ghost point layers. This modules must be placed in all routines which contain
!> MPI-calls.
!------------------------------------------------------------------------------!
 MODULE pegrid

    USE kinds

#if defined( __parallel )
#if defined( __mpifh )
    INCLUDE "mpif.h"
#else
    USE MPI
#endif
#endif
    CHARACTER(LEN=2) ::  send_receive = 'al'
    CHARACTER(LEN=7) ::  myid_char = ''
    
    INTEGER(iwp) ::  comm1dx                     !< 
    INTEGER(iwp) ::  comm1dy                     !< 
    INTEGER(iwp) ::  comm2d                      !< 
    INTEGER(iwp) ::  comm_inter                  !< 
    INTEGER(iwp) ::  comm_palm                   !< 
    INTEGER(iwp) ::  id_inflow = 0               !< 
    INTEGER(iwp) ::  id_outflow = 0              !< myidx of procs at outflow (turbulent outflow method)
    INTEGER(iwp) ::  id_outflow_source = 0       !< myidx of procs including ouflow source plane (turbulent outflow method)
    INTEGER(iwp) ::  id_recycling = 0            !< 
    INTEGER(iwp) ::  ierr                        !< 
    INTEGER(iwp) ::  myid = 0                    !< 
    INTEGER(iwp) ::  myidx = 0                   !< 
    INTEGER(iwp) ::  myidy = 0                   !< 
    INTEGER(iwp) ::  ndim = 2                    !< 
    INTEGER(iwp) ::  ngp_a                       !<
    INTEGER(iwp) ::  ngp_o                       !<
    INTEGER(iwp) ::  ngp_xy                      !< 
    INTEGER(iwp) ::  ngp_y                       !< 
    INTEGER(iwp) ::  npex = -1                   !< 
    INTEGER(iwp) ::  npey = -1                   !< 
    INTEGER(iwp) ::  numprocs = 1                !< 
    INTEGER(iwp) ::  numprocs_previous_run = -1  !< 
    INTEGER(iwp) ::  pleft                       !< 
    INTEGER(iwp) ::  pnorth                      !< 
    INTEGER(iwp) ::  pright                      !< 
    INTEGER(iwp) ::  psouth                      !< 
    INTEGER(iwp) ::  req_count = 0               !< 
    INTEGER(iwp) ::  sendrecvcount_xy            !< 
    INTEGER(iwp) ::  sendrecvcount_yz            !< 
    INTEGER(iwp) ::  sendrecvcount_zx            !< 
    INTEGER(iwp) ::  sendrecvcount_zyd           !< 
    INTEGER(iwp) ::  sendrecvcount_yxd           !< 
    INTEGER(iwp) ::  target_id                   !< 
    INTEGER(iwp) ::  tasks_per_node = -9999      !< 
    INTEGER(iwp) ::  threads_per_task = 1        !< 
    INTEGER(iwp) ::  type_x                      !< 
    INTEGER(iwp) ::  type_xy                     !< 
    INTEGER(iwp) ::  type_y                      !< 

    INTEGER(iwp) ::  pdims(2) = 1  !< 
    INTEGER(iwp) ::  req(100)      !< 

    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  hor_index_bounds               !< 
    INTEGER(iwp), DIMENSION(:,:), ALLOCATABLE ::  hor_index_bounds_previous_run  !< 

    LOGICAL ::  collective_wait = .FALSE.          !< 
    LOGICAL ::  sendrecv_in_background = .FALSE.   !< 

#if defined( __parallel )
#if defined( __mpi2 )
    CHARACTER (LEN=MPI_MAX_PORT_NAME) ::  port_name  !< 
#endif

    INTEGER(iwp) ::  ibuf(12)                 !< 
    INTEGER(iwp) ::  pcoord(2)                !< 
    INTEGER(iwp) ::  status(MPI_STATUS_SIZE)  !< 
    
    INTEGER(iwp), DIMENSION(MPI_STATUS_SIZE,100) ::  wait_stat  !< 
    
    INTEGER(iwp) ::  ngp_yz_int   !< 
    INTEGER(iwp) ::  type_xz_int  !< 
    INTEGER(iwp) ::  type_yz_int  !< 

    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  ngp_xz      !< 
    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  ngp_yz      !< 
    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  type_x_int  !< 
    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  type_xz     !< 
    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  type_y_int  !< 
    INTEGER(iwp), DIMENSION(:), ALLOCATABLE ::  type_yz     !< 

    LOGICAL ::  left_border_pe  = .FALSE.  !< 
    LOGICAL ::  north_border_pe = .FALSE.  !< 
    LOGICAL ::  reorder = .TRUE.           !< 
    LOGICAL ::  right_border_pe = .FALSE.  !< 
    LOGICAL ::  south_border_pe = .FALSE.  !< 

    LOGICAL, DIMENSION(2) ::  cyclic = (/ .TRUE. , .TRUE. /)  !< 
    LOGICAL, DIMENSION(2) ::  remain_dims                     !< 
#endif

    SAVE

 END MODULE pegrid


!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of variables which control PROFIL-output.
!------------------------------------------------------------------------------!
 MODULE profil_parameter

    USE kinds

    INTEGER(iwp), PARAMETER ::  crmax = 100  !< 

    CHARACTER (LEN=20), DIMENSION(20) ::  cross_ts_profiles = &  !< 
                           (/ ' E E*               ', ' dt                 ', &
                              ' u* w*              ', ' th*                ', &
                              ' umax vmax wmax     ', ' div_old div_new    ', &
                              ' z_i_wpt z_i_pt     ', ' w"pt"0 w"pt" wpt   ', &
                              ' pt(0) pt(zp)       ', ' splux spluy spluz  ', &
                              ' L                  ',                         &
                            ( '                    ', i9 = 1, 9 ) /)

    CHARACTER (LEN=100), DIMENSION(crmax) ::  cross_profiles = &  !< 
                           (/ ' u v                           ', &
                              ' pt                            ', &
                              ' w"pt" w*pt* w*pt*BC wpt wptBC ', &
                              ' w"u" w*u* wu w"v" w*v* wv     ', &
                              ' km kh                         ', &
                              ' l                             ', &
                         ( '                               ', i9 = 1, 94 ) /)

    INTEGER(iwp) ::  profile_columns = 2  !< 
    INTEGER(iwp) ::  profile_rows = 3     !< 
    INTEGER(iwp) ::  profile_number = 0   !< 

    INTEGER(iwp) ::  cross_ts_numbers(crmax,crmax) = 0  !< 
    INTEGER(iwp) ::  cross_ts_number_count(crmax) = 0   !< 
    INTEGER(iwp) ::  dopr_index(300) = 0                !< 
    INTEGER(iwp) ::  dopr_initial_index(300) = 0        !< 
    INTEGER(iwp) ::  dots_crossindex(100) = 0           !< 
    INTEGER(iwp) ::  dots_index(100) = 0                !< 
                

    REAL(wp) ::  cross_ts_uymax(20) = &                    !< 
                             (/ 999.999_wp, 999.999_wp, 999.999_wp, 999.999_wp, 999.999_wp,   &
                                999.999_wp, 999.999_wp, 999.999_wp, 999.999_wp, 999.999_wp,   &
                                999.999_wp, 999.999_wp, 999.999_wp, 999.999_wp, 999.999_wp,   &
                                999.999_wp, 999.999_wp, 999.999_wp, 999.999_wp, 999.999_wp /)
    REAL(wp) ::  cross_ts_uymax_computed(20) = 999.999_wp  !< 
    REAL(wp) ::  cross_ts_uymin(20) = &                    !< 
                             (/ 999.999_wp, 999.999_wp, 999.999_wp,  -5.000_wp, 999.999_wp,   &
                                999.999_wp,   0.000_wp, 999.999_wp, 999.999_wp, 999.999_wp,   &
                                999.999_wp, 999.999_wp, 999.999_wp, 999.999_wp, 999.999_wp,   &
                                999.999_wp, 999.999_wp, 999.999_wp, 999.999_wp, 999.999_wp /)
    REAL(wp) ::  cross_ts_uymin_computed(20) = 999.999_wp  !< 

    SAVE

 END MODULE profil_parameter

!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of statistical quantities, e.g. global sums.
!------------------------------------------------------------------------------!
 MODULE statistics

    USE kinds

    CHARACTER (LEN=40) ::  region(0:9)  !< 
    
    INTEGER(iwp) ::  pr_palm = 130          !< 
    INTEGER(iwp) ::  statistic_regions = 0  !< 
    
    INTEGER(iwp) ::  u_max_ijk(3) = -1  !< 
    INTEGER(iwp) ::  v_max_ijk(3) = -1  !< 
    INTEGER(iwp) ::  w_max_ijk(3) = -1  !< 
    
    LOGICAL ::  flow_statistics_called = .FALSE.  !< 
    
    REAL(wp) ::  u_max  !< 
    REAL(wp) ::  v_max  !< 
    REAL(wp) ::  w_max  !< 
    
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  mean_surface_level_height  !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  sums_divnew_l              !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  sums_divold_l              !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  weight_substep             !< 
    REAL(wp), DIMENSION(:), ALLOCATABLE ::  weight_pres                !< 
    
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums             !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_wsts_bc_l   !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  ts_value         !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_wsus_ws_l   !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_wsvs_ws_l   !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_us2_ws_l    !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_vs2_ws_l    !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_ws2_ws_l    !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_wsnrs_ws_l  !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_wspts_ws_l  !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_wsqs_ws_l   !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_wsqrs_ws_l  !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_wssas_ws_l  !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_wsss_ws_l   !< 
    REAL(wp), DIMENSION(:,:), ALLOCATABLE ::  sums_ls_l        !< 
                                              
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  hom_sum             !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  rmask               !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  sums_l              !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  sums_l_l            !< 
    REAL(wp), DIMENSION(:,:,:), ALLOCATABLE ::  sums_up_fraction_l  !< 

    REAL(wp), DIMENSION(:,:,:,:), ALLOCATABLE ::  hom  !< 

    SAVE

 END MODULE statistics



!------------------------------------------------------------------------------!
! Description:
! ------------
!> Definition of indices for transposed arrays.
!------------------------------------------------------------------------------!
 MODULE transpose_indices

    USE kinds

    INTEGER(iwp) ::  nxl_y   !< 
    INTEGER(iwp) ::  nxl_yd  !< 
    INTEGER(iwp) ::  nxl_z   !< 
    INTEGER(iwp) ::  nxr_y   !< 
    INTEGER(iwp) ::  nxr_yd  !< 
    INTEGER(iwp) ::  nxr_z   !< 
    INTEGER(iwp) ::  nyn_x   !< 
    INTEGER(iwp) ::  nyn_z   !< 
    INTEGER(iwp) ::  nys_x   !< 
    INTEGER(iwp) ::  nys_z   !< 
    INTEGER(iwp) ::  nzb_x   !< 
    INTEGER(iwp) ::  nzb_y   !< 
    INTEGER(iwp) ::  nzb_yd  !< 
    INTEGER(iwp) ::  nzt_x   !< 
    INTEGER(iwp) ::  nzt_y   !< 
    INTEGER(iwp) ::  nzt_yd  !< 
                
    SAVE

 END MODULE transpose_indices
